
/*global jQuery: false, GUI$: false, Cookies: false */

/*jslint vars: true */

/** 
* @preserve Copyright 2020 Pine Grove Software, LLC
* AccurateCalculators.com
* pine-grove.com
* interface.RETIRE-AGE-WIDGET.gpl.js
*/
(function ($, GUI) {
	'use strict';

	// don't try to initialize the wrong calculator
	if (!document.getElementById('retire-age-plugin')) {
		return;
	}


	var obj = {}, // interface object to base equations
		// gui controls
		currentAgeInput,
		pvInput,
		cfInput,
		rateInput,
		fvInput;


	/**
	* init() -- init or reset GUI's values
	*/
	function initGUI() {
		currentAgeInput.setValue(currentAgeInput.getUSNumber());
		pvInput.setValue(pvInput.getUSNumber());
		cfInput.setValue(cfInput.getUSNumber());
		rateInput.setValue(rateInput.getUSNumber());
		fvInput.setValue(fvInput.getUSNumber());

		document.getElementById("edRetirementAge-ra").value = GUI.formatLocalFloat(0, GUI.numConventions, 0);
		document.getElementById("edNumPmts-ra").value = GUI.formatLocalFloat(0, GUI.numConventions, 0);
		document.getElementById("edTotalInvested-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
		document.getElementById("edInterest-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
		document.getElementById("edFinalValue-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
		document.getElementById("edFVDate-ra").value = GUI.dateConventions.date_mask;

	} // initGUI



	/**
	* clearGUI() -- reset GUI's values
	*/
	function clearGUI() {

		currentAgeInput.setValue(0);
		pvInput.setValue(0.0);
		cfInput.setValue(0.0);
		rateInput.setValue(0.0);
		fvInput.setValue(0.0);

		document.getElementById("edRetirementAge-ra").value = GUI.formatLocalFloat(0, GUI.numConventions, 0);
		document.getElementById("edNumPmts-ra").value = GUI.formatLocalFloat(0, GUI.numConventions, 0);
		document.getElementById("edTotalInvested-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
		document.getElementById("edInterest-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
		document.getElementById("edFinalValue-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
		document.getElementById("edFVDate-ra").value = GUI.dateConventions.date_mask;

	} // clearGUI



	/**
	* getInputs() -- get user inputs and initialize obj equation interface object
	*/
	function getInputs() {
		// all rates are passed as decimal equivalents
		obj = {};
		obj.pv = pvInput.getUSNumber();
		obj.cf = cfInput.getUSNumber();
		obj.nominalRate = rateInput.getUSNumber() / 100;
		obj.fv = fvInput.getUSNumber();


		obj.oDate = GUI.dateMath.getFirstNextMonth(new Date());
		obj.oDate.setHours(0, 0, 0, 0);
		obj.fDate = GUI.dateMath.getFirstNextMonth(new Date(obj.oDate));
		obj.lDate = new Date();

		obj.pmtFreq = 6;
		obj.cmpFreq = 6;
		obj.pmtMthd = 0;

	} // getInputs()



	/** 
	* calc() -- initialize CashInputs data structures for equation classes
	*/
	function calc() {
		var invested, interest, bal, ca, ra;

		ca = currentAgeInput.getUSNumber();
		if (ca <= 0) {
			alert('Current age must be greater than 0.');
			return null;
		}
		if (obj.cf === 0 || obj.nominalRate === 0 || obj.fv === 0) {
			alert('There are unknown values.\nPlease make sure all values are entered.\n\n"Current Retirement Savings" can be 0.');
			return null;
		}

		obj.n = 0;
		obj.n = Math.ceil(GUI.N.calc(obj)); // round up
		ra = ca + (obj.n / 12);
		if (ra > 100) {
			alert('Your retirement age would be after age 100.\n\nIncrease investment amount.\n\nIncrease assumed rate of return.\n\nIncrease current retirement savings.\n\nOr do any combination of all three.');
			document.getElementById("edRetirementAge-ra").value = GUI.formatLocalFloat(0, GUI.numConventions, 0);
			document.getElementById("edNumPmts-ra").value = GUI.formatLocalFloat(0, GUI.numConventions, 0);
			document.getElementById("edTotalInvested-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
			document.getElementById("edInterest-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
			document.getElementById("edFinalValue-ra").value = GUI.formatLocalFloat(0.0, GUI.moneyConventions, 2);
			document.getElementById("edFVDate-ra").value = GUI.dateConventions.date_mask;
			return null;
		}

		obj.lDate.setTime(GUI.dateMath.addPeriods(obj.oDate, obj.n, obj.pmtFreq));
		GUI.RETIRE_AGE_SCHEDULE.calc(obj);
		invested = GUI.summary.totalPmts[1];
		interest = GUI.summary.totalInterest[1];
		bal = GUI.summary.unadjustedBalance;

		document.getElementById("edRetirementAge-ra").value = GUI.formatLocalFloat(GUI.roundMoney(ra), GUI.numConventions, 0);
		document.getElementById("edNumPmts-ra").value = GUI.formatLocalFloat(obj.n, GUI.numConventions, 0);
		document.getElementById("edTotalInvested-ra").value = GUI.formatLocalFloat(GUI.roundMoney(invested, 2), GUI.moneyConventions, 2);
		document.getElementById("edInterest-ra").value = GUI.formatLocalFloat(GUI.roundMoney(interest, 2), GUI.moneyConventions, 2);
		document.getElementById("edFinalValue-ra").value = GUI.formatLocalFloat(GUI.roundMoney(bal, 2), GUI.moneyConventions, 2);
		document.getElementById("edFVDate-ra").value = GUI.dateMath.dateToDateStr(obj.lDate, GUI.dateConventions);
		return 1;
	} // function calc()




	$(document).ready(function () {

		// If user has a ccy_format cookie, use its value first
		// otherwise if website set default currency use it
		// otherwise use currency determined by user's locale - previously initialized
		var currency = parseInt(document.getElementById('currency-ra').value, 10);
		if (Cookies.get('ccy_format')) {
			GUI.updateNumConventions(parseInt(Cookies.get('ccy_format'), 10));
			// GUI.moneyConventions = new GUI.LocalConventions(parseInt(Cookies.get('ccy_format'), 10));
			// // clones moneyConventions and sets ccy_r = '%'
			// GUI.rateConventions = GUI.moneyConventions.rateConvention(); // clones currency conventions with'%' symbol
			// GUI.numConventions = GUI.moneyConventions.numConvention(); // clones currency conventions without currency
		} else if (currency !== undefined && currency !== null && typeof currency === 'number' && currency !== 999) {
			GUI.updateNumConventions(currency);
		}

		// If user has a date_format cookie, use its value first
		// otherwise if website set default date_mask use it
		// otherwise use date_mask determined by user's locale - previously initialized
		var date_mask = parseInt(document.getElementById('date_mask-ra').value, 10);
		if (Cookies.get('date_format')) {
			GUI.dateConventions = new GUI.LocalDateConventions(parseInt(Cookies.get('date_format'), 10));
		} else if (date_mask !== undefined && date_mask !== null && typeof date_mask === 'number' && date_mask !== 999) {
			GUI.updateDateConventions(date_mask);
		}

		//
		// initialize GUI controls & dialog / modal controls here
		// attach
		//
		// only required on the AccurateCalculators.com site
		// here for each comment reply link of WordPress
		// $('.comment-reply-link').addClass('btn btn-primary');

		// here for the submit button of the comment reply form
		// $('#submit').addClass('btn btn-primary');

		// Style contact form submit button
		// $('.wpcf7-submit').addClass('btn btn-primary');

		// Add thumbnail styling
		// $('.wp-caption').addClass('thumbnail');

		// Now we'll add some classes for the WordPress default widgets - let's go
		// Add Bootstrap style for drop-downs
		// $('.postform').addClass('form-control');
		// end: only required on the AccurateCalculators.com site

		// main window
		currentAgeInput = new GUI.NE('edCurrentAge-ra', GUI.numConventions, 0);
		pvInput = new GUI.NE('edPV-ra', GUI.moneyConventions, 2);
		cfInput = new GUI.NE('edCF-ra', GUI.moneyConventions, 2);
		rateInput = new GUI.NE('edRate-ra', GUI.rateConventions, 4);
		fvInput = new GUI.NE('edFV-ra', GUI.moneyConventions, 2);

		initGUI();


		$('#btnCalc-ra').click(function () {
			getInputs();
			calc();
		});


		$('#btnClear-ra').click(function () {
			clearGUI();
		});


		$('#btnPrint-ra').click(function () {
			getInputs();
			calc();
			GUI.print_calc();
		});


		$('#btnSchedule-ra').click(function () {
			getInputs();
			if (calc() !== null) {
				GUI.showSavingsSchedule(GUI.RETIRE_AGE_SCHEDULE.calc(obj));
			}
		});


		$('#btnCharts-ra').click(function () {
			getInputs();
			if (calc() !== null) {
				GUI.showSavingsCharts(GUI.RETIRE_AGE_SCHEDULE.calc(obj));
			}
		});


		$('#btnHelp-ra').click(function () {
			GUI.show_help('#hText-retire-age');
		});


//		$('#btnCcyDate, #btnCcyDate2, #CCY').click(function () {
		$('#btnCcyDate, #btnCcyDate2, #CCY-ra').click(function () {
			GUI$.init_CURRENCYDATE_Dlg();
		});

	}); // $(document).ready

}(jQuery, GUI$));
